@@ -76,9 +76,6 @@ module RuboCop
       OPERATOR_KEYWORDS = %i[and or].to_set.freeze
       # @api private
       SPECIAL_KEYWORDS = %w[__FILE__ __LINE__ __ENCODING__].to_set.freeze
-      # @api private
-      ARGUMENT_TYPES = %i[arg optarg restarg kwarg kwoptarg kwrestarg
-                          blockarg forward_arg shadowarg].to_set.freeze
 
       LITERAL_RECURSIVE_METHODS = (COMPARISON_OPERATORS + %i[* ! <=>]).freeze
       LITERAL_RECURSIVE_TYPES = (OPERATOR_KEYWORDS + COMPOSITE_LITERALS + %i[begin pair]).freeze
@@ -88,6 +85,59 @@ module RuboCop
       EMPTY_PROPERTIES = {}.freeze
       private_constant :EMPTY_CHILDREN, :EMPTY_PROPERTIES
 
+      # @api private
+      GROUP_FOR_TYPE = {
+        arg: :argument,
+        optarg: :argument,
+        restarg: :argument,
+        kwarg: :argument,
+        kwoptarg: :argument,
+        kwrestarg: :argument,
+        blockarg: :argument,
+        forward_arg: :argument,
+        shadowarg: :argument,
+
+        true: :boolean,
+        false: :boolean,
+
+        int: :numeric,
+        float: :numeric,
+        rational: :numeric,
+        complex: :numeric,
+
+        irange: :range,
+        erange: :range,
+
+        send: :call,
+        csend: :call,
+
+        block: :any_block,
+        numblock: :any_block,
+        itblock: :any_block
+      }.freeze
+      private_constant :GROUP_FOR_TYPE
+
+      # Define a +recursive_?+ predicate method for the given node kind.
+      private_class_method def self.def_recursive_literal_predicate(kind) # rubocop:disable Metrics/MethodLength
+        recursive_kind = "recursive_#{kind}?"
+        kind_filter = "#{kind}?"
+
+        class_eval <<~RUBY, __FILE__, __LINE__ + 1
+          def #{recursive_kind}                                     # def recursive_literal?
+            case type                                               #   case type
+            when :send                                              #   when :send
+              LITERAL_RECURSIVE_METHODS.include?(method_name) &&    #     LITERAL_RECURSIVE_METHODS.include?(method_name) &&
+                receiver.send(:#{recursive_kind}) &&                #       receiver.send(:recursive_literal?) &&
+                arguments.all?(&:#{recursive_kind})                 #       arguments.all?(&:recursive_literal?)
+            when LITERAL_RECURSIVE_TYPES                            #   when LITERAL_RECURSIVE_TYPES
+              children.compact.all?(&:#{recursive_kind})            #     children.compact.all?(&:recursive_literal?)
+            else                                                    #   else
+              send(:#{kind_filter})                                 #     send(:literal?)
+            end                                                     #   end
+          end                                                       # end
+        RUBY
+      end
+
       # @see https://www.rubydoc.info/gems/ast/AST/Node:initialize
       def initialize(type, children = EMPTY_CHILDREN, properties = EMPTY_PROPERTIES)
         @mutable_attributes = {}
@@ -105,6 +155,16 @@ module RuboCop
         end
       end
 
+      # Determine if the node is one of several node types in a single query
+      # Allows specific single node types, as well as "grouped" types
+      # (e.g. `:boolean` for `:true` or `:false`)
+      def type?(*types)
+        return true if types.include?(type)
+
+        group_type = GROUP_FOR_TYPE[type]
+        !group_type.nil? && types.include?(group_type)
+      end
+
       (Parser::Meta::NODE_TYPES - [:send]).each do |node_type|
         method_name = "#{node_type.to_s.gsub(/\W/, '')}_type?"
         class_eval <<~RUBY, __FILE__, __LINE__ + 1
@@ -284,20 +344,19 @@ module RuboCop
 
       # @!method receiver(node = self)
       def_node_matcher :receiver, <<~PATTERN
-        {(send $_ ...) ({block numblock} (call $_ ...) ...)}
+        {(send $_ ...) (any_block (call $_ ...) ...)}
       PATTERN
 
       # @!method str_content(node = self)
       def_node_matcher :str_content, '(str $_)'
 
       def const_name
-        return unless const_type?
+        return unless const_type? || casgn_type?
 
-        namespace, name = *self
         if namespace && !namespace.cbase_type?
-          "#{namespace.const_name}::#{name}"
+          "#{namespace.const_name}::#{short_name}"
         else
-          name.to_s
+          short_name.to_s
         end
       end
 
@@ -326,13 +385,13 @@ module RuboCop
         # what class or module is this method/constant/etc definition in?
         # returns nil if answer cannot be determined
         ancestors = each_ancestor(:class, :module, :sclass, :casgn, :block)
-        result    = ancestors.map do |ancestor|
+        result    = ancestors.filter_map do |ancestor|
           parent_module_name_part(ancestor) do |full_name|
             return nil unless full_name
 
             full_name
           end
-        end.compact.reverse.join('::')
+        end.reverse.join('::')
         result.empty? ? 'Object' : result
       end
 
@@ -380,22 +439,11 @@ module RuboCop
         IMMUTABLE_LITERALS.include?(type)
       end
 
-      %i[literal basic_literal].each do |kind|
-        recursive_kind = :"recursive_#{kind}?"
-        kind_filter = :"#{kind}?"
-        define_method(recursive_kind) do
-          case type
-          when :send
-            LITERAL_RECURSIVE_METHODS.include?(method_name) &&
-              receiver.send(recursive_kind) &&
-              arguments.all?(&recursive_kind)
-          when LITERAL_RECURSIVE_TYPES
-            children.compact.all?(&recursive_kind)
-          else
-            send(kind_filter)
-          end
-        end
-      end
+      # @!macro [attach] def_recursive_literal_predicate
+      #   @!method recursive_$1?
+      #     @return [Boolean]
+      def_recursive_literal_predicate :literal
+      def_recursive_literal_predicate :basic_literal
 
       def variable?
         VARIABLES.include?(type)
@@ -450,11 +498,11 @@ module RuboCop
       end
 
       def parenthesized_call?
-        loc.respond_to?(:begin) && loc.begin && loc.begin.is?('(')
+        loc_is?(:begin, '(')
       end
 
       def call_type?
-        send_type? || csend_type?
+        GROUP_FOR_TYPE[type] == :call
       end
 
       def chained?
@@ -466,27 +514,47 @@ module RuboCop
       end
 
       def argument_type?
-        ARGUMENT_TYPES.include?(type)
+        GROUP_FOR_TYPE[type] == :argument
       end
 
       def boolean_type?
-        true_type? || false_type?
+        GROUP_FOR_TYPE[type] == :boolean
       end
 
       def numeric_type?
-        int_type? || float_type? || rational_type? || complex_type?
+        GROUP_FOR_TYPE[type] == :numeric
       end
 
       def range_type?
-        irange_type? || erange_type?
+        GROUP_FOR_TYPE[type] == :range
+      end
+
+      def any_block_type?
+        GROUP_FOR_TYPE[type] == :any_block
       end
 
       def guard_clause?
-        node = and_type? || or_type? ? rhs : self
+        node = operator_keyword? ? rhs : self
 
         node.match_guard_clause?
       end
 
+      # Shortcut to safely check if a location is present
+      # @return [Boolean]
+      def loc?(which_loc)
+        return false unless loc.respond_to?(which_loc)
+
+        !loc.public_send(which_loc).nil?
+      end
+
+      # Shortcut to safely test a particular location, even if
+      # this location does not exist or is `nil`
+      def loc_is?(which_loc, str)
+        return false unless loc?(which_loc)
+
+        loc.public_send(which_loc).is?(str)
+      end
+
       # @!method match_guard_clause?(node = self)
       def_node_matcher :match_guard_clause?, <<~PATTERN
         [${(send nil? {:raise :fail} ...) return break next} single_line?]
@@ -500,7 +568,7 @@ module RuboCop
       PATTERN
 
       # @!method lambda?(node = self)
-      def_node_matcher :lambda?, '({block numblock} (send nil? :lambda) ...)'
+      def_node_matcher :lambda?, '(any_block (send nil? :lambda) ...)'
 
       # @!method lambda_or_proc?(node = self)
       def_node_matcher :lambda_or_proc?, '{lambda? proc?}'
@@ -513,7 +581,7 @@ module RuboCop
         {
           (send #global_const?({:Class :Module :Struct}) :new ...)
           (send #global_const?(:Data) :define ...)
-          ({block numblock} {
+          (any_block {
             (send #global_const?({:Class :Module :Struct}) :new ...)
             (send #global_const?(:Data) :define ...)
           } ...)
@@ -523,20 +591,20 @@ module RuboCop
       # @deprecated Use `:class_constructor?`
       # @!method struct_constructor?(node = self)
       def_node_matcher :struct_constructor?, <<~PATTERN
-        ({block numblock} (send #global_const?(:Struct) :new ...) _ $_)
+        (any_block (send #global_const?(:Struct) :new ...) _ $_)
       PATTERN
 
       # @!method class_definition?(node = self)
       def_node_matcher :class_definition?, <<~PATTERN
         {(class _ _ $_)
          (sclass _ $_)
-         ({block numblock} (send #global_const?({:Struct :Class}) :new ...) _ $_)}
+         (any_block (send #global_const?({:Struct :Class}) :new ...) _ $_)}
       PATTERN
 
       # @!method module_definition?(node = self)
       def_node_matcher :module_definition?, <<~PATTERN
         {(module _ $_)
-         ({block numblock} (send #global_const?(:Module) :new ...) _ $_)}
+         (any_block (send #global_const?(:Module) :new ...) _ $_)}
       PATTERN
 
       # Some expressions are evaluated for their value, some for their side
@@ -547,8 +615,7 @@ module RuboCop
       # So, does the return value of this node matter? If we changed it to
       # `(...; nil)`, might that affect anything?
       #
-      # rubocop:disable Metrics/MethodLength
-      def value_used?
+      def value_used? # rubocop:disable Metrics/MethodLength
         # Be conservative and return true if we're not sure.
         return false if parent.nil?
 
@@ -569,7 +636,6 @@ module RuboCop
           true
         end
       end
-      # rubocop:enable Metrics/MethodLength
 
       # Some expressions are evaluated for their value, some for their side
       # effects, and some for both.
@@ -601,8 +667,7 @@ module RuboCop
         last_node = self
 
         while (current_node = last_node.parent)
-          yield current_node if types.empty? ||
-                                types.include?(current_node.type)
+          yield current_node if types.empty? || current_node.type?(*types)
           last_node = current_node
         end
       end
@@ -621,7 +686,7 @@ module RuboCop
       def case_if_value_used?
         # (case <condition> <when...>)
         # (if <condition> <truebranch> <falsebranch>)
-        sibling_index.zero? ? true : parent.value_used?
+        sibling_index.zero? || parent.value_used?
       end
 
       def while_until_value_used?
